@@ -9,6 +9,8 @@ This module provides data structures and utilities common
 to all implementations of OAuth.
 """
 
+import collections
+import logging
 import random
 import re
 import sys
@@ -30,12 +32,29 @@ UNICODE_ASCII_CHARACTER_SET = ('abcdefghijklmnopqrstuvwxyz'
                                'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                                '0123456789')
 
+CLIENT_ID_CHARACTER_SET = (r' !"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMN'
+                            'OPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}')
+
+
 always_safe = ('ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                'abcdefghijklmnopqrstuvwxyz'
                '0123456789' '_.-')
 
 PY3 = sys.version_info[0] == 3
 
+# Logger used throughout oauthlib
+log = logging.getLogger('oauthlib')
+# Add a NullHandler to prevent warnings for users who don't wish
+# to configure logging.
+try:
+    log.addHandler(logging.NullHandler())
+# NullHandler gracefully backported to 2.6
+except AttributeError:
+    class NullHandler(logging.Handler):
+
+        def emit(self, record):
+            pass
+    log.addHandler(NullHandler())
 
 if PY3:
     unicode_type = str
@@ -47,7 +66,8 @@ else:
 
 # 'safe' must be bytes (Python 2.6 requires bytes, other versions allow either)
 def quote(s, safe=b'/'):
-    s = _quote(s.encode('utf-8'), safe)
+    s = s.encode('utf-8') if isinstance(s, unicode_type) else s
+    s = _quote(s, safe)
     # PY3 always returns unicode.  PY2 may return either, depending on whether
     # it had to modify the string.
     if isinstance(s, bytes_type):
@@ -98,7 +118,7 @@ def decode_params_utf8(params):
     return decoded
 
 
-urlencoded = set(always_safe) | set('=&;%+~')
+urlencoded = set(always_safe) | set('=&;%+~,*')
 
 
 def urldecode(query):
@@ -122,7 +142,21 @@ def urldecode(query):
     if len(re.findall(invalid_hex, query)):
         raise ValueError('Invalid hex encoding in query string.')
 
-    query = query.decode('utf-8') if isinstance(query, bytes_type) else query
+    # We encode to utf-8 prior to parsing because parse_qsl behaves
+    # differently on unicode input in python 2 and 3.
+    # Python 2.7
+    # >>> urlparse.parse_qsl(u'%E5%95%A6%E5%95%A6')
+    # u'\xe5\x95\xa6\xe5\x95\xa6'
+    # Python 2.7, non unicode input gives the same
+    # >>> urlparse.parse_qsl('%E5%95%A6%E5%95%A6')
+    # '\xe5\x95\xa6\xe5\x95\xa6'
+    # but now we can decode it to unicode
+    # >>> urlparse.parse_qsl('%E5%95%A6%E5%95%A6').decode('utf-8')
+    # u'\u5566\u5566'
+    # Python 3.3 however
+    # >>> urllib.parse.parse_qsl(u'%E5%95%A6%E5%95%A6')
+    # u'\u5566\u5566'
+    query = query.encode('utf-8') if not PY3 and isinstance(query, unicode_type) else query
     # We want to allow queries such as "c2" whereas urlparse.parse_qsl
     # with the strict_parsing flag will not.
     params = urlparse.parse_qsl(query, keep_blank_values=True)
@@ -199,6 +233,15 @@ def generate_token(length=30, chars=UNICODE_ASCII_CHARACTER_SET):
     return ''.join(rand.choice(chars) for x in range(length))
 
 
+def generate_client_id(length=30, chars=CLIENT_ID_CHARACTER_SET):
+    """Generates an OAuth client_id
+
+    OAuth 2 specify the format of client_id in
+    http://tools.ietf.org/html/rfc6749#appendix-A.
+    """
+    return generate_token(length, chars)
+
+
 def add_params_to_qs(query, params):
     """Extend a query with a list of two-tuples."""
     if isinstance(query, dict):
@@ -255,13 +298,43 @@ def to_unicode(data, encoding):
             return (to_unicode(i, encoding) for i in data)
         else:
             # We support 2.6 which lacks dict comprehensions
-            if isinstance(data, dict):
+            if hasattr(data, 'items'):
                 data = data.items()
             return dict(((to_unicode(k, encoding), to_unicode(v, encoding)) for k, v in data))
 
     return data
 
 
+class CaseInsensitiveDict(dict):
+    """Basic case insensitive dict with strings only keys."""
+
+    proxy = {}
+
+    def __init__(self, data):
+        self.proxy = dict((k.lower(), k) for k in data)
+        for k in data:
+            self[k] = data[k]
+
+    def __contains__(self, k):
+        return k.lower() in self.proxy
+
+    def __delitem__(self, k):
+        key = self.proxy[k.lower()]
+        super(CaseInsensitiveDict, self).__delitem__(key)
+        del self.proxy[k.lower()]
+
+    def __getitem__(self, k):
+        key = self.proxy[k.lower()]
+        return super(CaseInsensitiveDict, self).__getitem__(key)
+
+    def get(self, k, default=None):
+        return self[k] if k in self else default
+
+    def __setitem__(self, k, v):
+        super(CaseInsensitiveDict, self).__setitem__(k, v)
+        self.proxy[k.lower()] = k
+
+
 class Request(object):
     """A malleable representation of a signable HTTP request.
 
@@ -283,7 +356,7 @@ class Request(object):
 
         self.uri = encode(uri)
         self.http_method = encode(http_method)
-        self.headers = encode(headers or {})
+        self.headers = CaseInsensitiveDict(encode(headers or {}))
         self.body = encode(body)
         self.decoded_body = extract_params(encode(body))
         self.oauth_params = []
@@ -302,5 +375,15 @@ class Request(object):
 
     @property
     def uri_query_params(self):
+        if not self.uri_query:
+            return []
         return urlparse.parse_qsl(self.uri_query, keep_blank_values=True,
                                   strict_parsing=True)
+
+    @property
+    def duplicate_params(self):
+        seen_keys = collections.defaultdict(int)
+        all_keys = (p[0] for p in self.decoded_body or [] + self.uri_query_params)
+        for k in all_keys:
+            seen_keys[k] += 1
+        return [k for k, c in seen_keys.items() if c > 1]
